#region using
using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
#endregion

namespace Cotyounoyume{
	public class ExpressionBlushingAndTears : MVRScript
	{
		#region common parameters
		protected Atom atom;
		protected string myClassName = "ExpressionBlushingAndTears";
		protected string myPath;
		protected float openingTime;
		protected float currentTime;
		protected int gauge = 0;
		protected JSONStorableBool atOnce;
		protected SimpleJSON.JSONClass settings = new SimpleJSON.JSONClass();
		protected string defaultSettingsPathBase = "Config/defaultSettings.json";
		protected string defaultSettingsPath;
		protected string currentSettingsPath = "Custom/Scripts/ExpressionBlushingAndTears_currentSettings.json";

		protected string pluginLabel;
		protected bool autoExpressionOffGlobal = true;
		protected bool lookModeGlobal = true;		
		protected bool tearsOnGlobal = true;
		protected bool droolOnGlobal = true;
		protected bool blushOnGlobal = true;
		protected bool expressionOnGlobal = true;
		protected bool eyecontrolOnGlobal = true;
		protected bool eyecontrolRollupOnGlobal = true;
		protected bool isTearsGloss = true;
		protected bool disableLog = true;
		protected int[] stageThreshold;
		#endregion

		#region blush parameters
		protected float blushClosingTimeOffset = 10f;
		protected float blushClosingTime;
		protected JSONStorableUrl faceDecalUrl;
		protected bool isBlushingActive;
		protected bool blushLevelUp = false;
		protected string blushTypeDefault = "Cartoonlike-Large-Alt";
		protected string blushAliasLow = "";
		protected string blushAliasHigh = "";
		Dictionary<string, UIDynamicToggle> toggles = new Dictionary<string, UIDynamicToggle>();
		List<string> blushTypes;
		protected Dictionary<string, string> blushMap;
		#endregion
		
		#region expression parameters
		protected float expressionClosingTimeOffset1 = 0.2f;
		protected float expressionClosingTimeOffset2 = 9.8f;
		protected float expressionOpeningTime;
		protected GenerateDAZMorphsControlUI morphControl;
		protected DAZMorph[] Emorph;
		protected DAZMorph EmorphEyesClosed;
		protected bool isExpressionActive;
		protected float targetExpressionStrength;
		protected float initialExpressionStrength;
		protected float currentExpressionStrength;
		protected float targetEyeClose;
		protected float initialEyeClose;
		protected float currentEyeClose;
		protected int expressionPattern = 0;
		protected int expressionPatternOld = 0;
		protected float[] eyeCloseMax;

		protected float expressionRandomOffset = 1.0f;
		#endregion
		
		#region tears parameters
		protected float tearsClosingTime;
		protected float tearsClosingTimeOffset = 10f;
		protected DAZCharacterSelector geometryForClothes;
		protected Dictionary<string, DAZClothingItem> tearsMap;
		protected bool isTearsUpdated;
		protected string tearsString;
		protected string droolString;
		protected bool isTearsActive = false;
		protected bool droolOn = false;
		protected bool tearsOn = false;		
		#endregion
		
		#region eyeControl
		protected Transform eyeControlOrig;
		protected FreeControllerV3 headControl;
		protected FreeControllerV3 eyeControl;
		protected float eyecontrolDegree;
		protected float baseDegree = 30f;
		protected float eyecontrolClosingTime;
		protected float eyecontrolClosingTimeInterval = 2f;
		protected float eyecontrolOpeningTime;
		protected float eyecontrolRandomOffset = 0.2f;
		protected bool isEyecontrolActive = false;
		#endregion
		
		#region ui var
		protected JSONStorableStringChooser blushType;
		protected JSONStorableBool toggleAE;
		protected JSONStorableBool toggleBlush;
		protected JSONStorableBool toggleTears;
		protected JSONStorableBool toggleDrool;
		protected JSONStorableBool toggleTearsGloss;
		protected JSONStorableBool toggleExpression;
		protected JSONStorableBool toggleEyecontrol;
		protected JSONStorableBool toggleEyecontrolRollup;
		protected JSONStorableBool toggleDebug;
		
		protected JSONStorableFloat jfBlush;
		protected JSONStorableFloat jfTears;
		protected JSONStorableFloat jfEyecontrol;
		protected JSONStorableFloat jfExpressionUp;
		protected JSONStorableFloat jfExpressionDown;
		
		protected JSONStorableFloat jfThreshold1;
		protected JSONStorableFloat jfThreshold2;
		protected JSONStorableFloat jfThreshold3;
		protected JSONStorableFloat jfThreshold4;
		protected UIDynamicButton resetButton;		
		#endregion

		public override void Init(){
			try{
				if(atOnce == null){
					logger("Plugin installed");
					atOnce = new JSONStorableBool("atOnce", false);
				}else{
					logger("already loaded.");
				}
                if (containingAtom.type != "Person")
                {
                    SuperController.LogError("This plug-in have to be added to person type atom.");
                    return;
                }
				atom = containingAtom;
				logger("atom name: " + atom.name);
				myPath = GetPluginPath();
				defaultSettingsPath = GetPluginPath() + "/" + defaultSettingsPathBase;
			}catch (Exception e){
				SuperController.LogError("Failed to initialize plugin: " + e);
			}
		}

		public void Start(){
			#region at onece: base
			if(configFileExist() == false){
				loadSettings("default");
				saveConfig();
			}
			loadSettings("current");
			applySettings();
			setupUI();
			openingTime = Time.time;
			logger("openingTime = " + openingTime);
			logger("myPath:" + myPath);
			gauge = 0;
			isExpressionActive = false;
			isBlushingActive = false;
			isTearsActive = false;
			#endregion

			maintenanceSequence();
		}

		public bool configFileExist(){
			JSONNode js = SuperController.singleton.LoadJSON(currentSettingsPath);
			if(js == null){
				SuperController.LogMessage(currentSettingsPath + " doesn't exist.");
				return false;
			}else{
				logger(currentSettingsPath + " exists.");
				return true;
			}
		}
		
		public void loadSettings(string type){
			string path = (type == "default") ? defaultSettingsPath : currentSettingsPath;
			settings = SuperController.singleton.LoadJSON(path) as JSONClass;
			
			#region general settings from json
			autoExpressionOffGlobal = settings["General"]["AutoExpressionsOff"].AsBool;
			lookModeGlobal = settings["General"]["LookMode"].AsBool;
			pluginLabel = settings["General"]["Label"];
			blushOnGlobal = settings["General"]["BlushOn"].AsBool;
			tearsOnGlobal = settings["General"]["TearsOn"].AsBool;
			droolOnGlobal = settings["General"]["DroolOn"].AsBool;
			isTearsGloss = settings["General"]["TearsGlossOn"].AsBool;
			expressionOnGlobal = settings["General"]["ExpressionOn"].AsBool;
			eyecontrolOnGlobal = settings["General"]["EyesControlOn"].AsBool;
			eyecontrolRollupOnGlobal = settings["General"]["EyesRollupOn"].AsBool;
			blushTypeDefault = settings["General"]["DefaultBlushType"];
			disableLog = settings["General"]["DebugMessageOff"].AsBool;
			stageThreshold = new int[5];
			stageThreshold[0] = 0;
			for(int i = 1; i <= 4; i++){
				stageThreshold[i] = settings["General"]["StageThreshold" + i].AsInt;				
			}
			blushClosingTimeOffset = settings["General"]["BlushDurationTime"].AsFloat;
			tearsClosingTimeOffset = settings["General"]["TearsDurationTime"].AsFloat;
			expressionClosingTimeOffset1 = settings["General"]["ExpressionUpTime"].AsFloat;
			expressionClosingTimeOffset2 = settings["General"]["ExpressionDownTime"].AsFloat;
			eyecontrolClosingTimeInterval = settings["General"]["EyesControlDurationTime"].AsFloat;
			#endregion
			
			#region read expression from json
			JSONNode jsEmorph = settings["Emorph"];
			JSONNode jsEyeCloseMax = settings["EyeCloseMax"];
			int num = jsEmorph.Count;
			
			Emorph = new DAZMorph[num];
			eyeCloseMax = new float[num];
			JSONStorable geometry = atom.GetStorableByID("geometry");
            DAZCharacterSelector character = geometry as DAZCharacterSelector;
            morphControl = character.morphsControlUI;
			
			for(int i = 0; i < num; i++){
				Emorph[i] = morphControl.GetMorphByDisplayName(jsEmorph[i]);
				eyeCloseMax[i] = jsEyeCloseMax[i].AsFloat;
			}
			EmorphEyesClosed = morphControl.GetMorphByDisplayName("Eyes Closed");
			#endregion
			
			#region read blushing from json
			blushTypes = new List<string>();
			blushMap = new Dictionary<string, string>();
			JSONNode jsBlush = settings["Blush"];
			for(int i = 0; i < jsBlush.Count; i++){
				JSONNode jsPair = jsBlush[i];
				blushTypes.Add(jsPair["Name"]);
				string fullpath = myPath + "/Blushing/" + jsPair["File"];
				blushMap.Add(jsPair["Name"], fullpath);
			}
			#endregion
		}
		
		public void applySettings(){
			pluginLabelJSON.val = settings["General"]["Label"];
			blushAliasLow = blushMap[blushTypeDefault];
			atom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", !(settings["General"]["AutoExpressionsOff"].AsBool));
			atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", settings["General"]["LookMode"]);
		}
		
		public void maintenanceSequence(){
			#region initialize eyeControl
			headControl = atom.GetStorableByID("headControl") as FreeControllerV3;
			eyeControl = atom.GetStorableByID("eyeTargetControl") as FreeControllerV3;			
			setInitialTransformOfEyeControl();
			#endregion
		
			#region initialize blush
			JSONStorable receiver = atom.GetStorableByID("textures");
			faceDecalUrl = receiver.GetUrlJSONParam("faceDecalUrl");
			#endregion
			
			#region initialize trigger
			resetTriggers();
			RegisterActions();
			#endregion
			
			#region initialize tears(clothes)
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
            geometryForClothes = geometryStorable as DAZCharacterSelector;
			
			tearsMap = new Dictionary<string, DAZClothingItem>();			
			List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
			for (int i = 0; i < clothes.Count; i++){
				if(clothes[i].creatorName == "cotyounoyume"){
					logger("clothes: " + clothes[i].creatorName + "/" + clothes[i].displayName);
					tearsMap.Add(clothes[i].displayName, clothes[i]);
				}
            }
			#endregion
		}
		
		public void setupUI(){
			 toggleAE = new JSONStorableBool("Auto Expression Off", autoExpressionOffGlobal, (bool toggle) => {
				autoExpressionOffGlobal = autoExpressionOffGlobal ? false : true;
				settings["General"]["AutoExpressionsOff"] = autoExpressionOffGlobal ? "true" : "false";
				atom.GetStorableByID("AutoExpressions").SetBoolParamValue("enabled", !(autoExpressionOffGlobal));
				saveConfig();
			});
			CreateToggle((toggleAE), false);
			
			 toggleBlush = new JSONStorableBool("Blush On", blushOnGlobal, (bool toggle) => {
				blushOnGlobal = blushOnGlobal ? false : true;
				settings["General"]["BlushOn"] = blushOnGlobal ? "true" : "false";
				if(blushOnGlobal){
					updateBlush();
				}else{
					isBlushingActive = false;
					switchFaceDecal("");
				}
				saveConfig();
			});
			CreateToggle((toggleBlush), false);
			
			 toggleTears = new JSONStorableBool("Tears On", tearsOnGlobal, (bool toggle) => {
				tearsOnGlobal = tearsOnGlobal ? false : true;
				settings["General"]["TearsOn"] = tearsOnGlobal ? "true" : "false";
				saveConfig();
				refreshTears();
			});
			CreateToggle((toggleTears), false);
			
			 toggleDrool = new JSONStorableBool("Drool On", droolOnGlobal, (bool toggle) => {
				droolOnGlobal = droolOnGlobal ? false : true;
				settings["General"]["DroolOn"] = droolOnGlobal ? "true" : "false";
				saveConfig();
				refreshTears();
			});
			CreateToggle((toggleDrool), false);
			
			 toggleTearsGloss = new JSONStorableBool("Tears Gloss On", isTearsGloss, (bool toggle) => {
				isTearsGloss = isTearsGloss ? false : true;
				settings["General"]["TearsGlossOn"] = isTearsGloss ? "true" : "false";
				saveConfig();
				refreshTears();
			});
			CreateToggle((toggleTearsGloss), false);
			
			 toggleExpression = new JSONStorableBool("Expression On", expressionOnGlobal, (bool toggle) => {
				expressionOnGlobal = expressionOnGlobal ? false : true;
				settings["General"]["ExpressionOn"] = expressionOnGlobal ? "true" : "false";
				if(expressionOnGlobal){
					updateExpression();
				}else{
					stopExpression();
					currentExpressionStrength = 0f;
				}
				saveConfig();
			});
			CreateToggle((toggleExpression), false);
			
			 toggleEyecontrol = new JSONStorableBool("Eyes Control On", eyecontrolOnGlobal, (bool toggle) => {
				eyecontrolOnGlobal = eyecontrolOnGlobal ? false : true;
				settings["General"]["EyesControlOn"] = eyecontrolOnGlobal ? "true" : "false";
				if(eyecontrolOnGlobal){
					updateEyecontrol();
				}else{
					eyeControlBaseSwitch("Player");
					setInitialTransformOfEyeControl();
					isEyecontrolActive = false;
				}
				saveConfig();
			});
			CreateToggle((toggleEyecontrol), false);
			
			 toggleEyecontrolRollup = new JSONStorableBool("Eyes Rollup On", eyecontrolRollupOnGlobal, (bool toggle) => {
				eyecontrolRollupOnGlobal = eyecontrolRollupOnGlobal ? false : true;
				settings["General"]["EyesRollupOn"] = eyecontrolRollupOnGlobal ? "true" : "false";
				saveConfig();
			});
			CreateToggle((toggleEyecontrolRollup), false);
			
			 jfBlush = new JSONStorableFloat("Blush Duration Time", blushClosingTimeOffset, (float value) => {
				blushClosingTimeOffset = value;
				settings["General"]["BlushDurationTime"] = value.ToString();
				saveConfig();
			}, 0f, 60f);
			CreateSlider(jfBlush, true);

			 jfTears = new JSONStorableFloat("Tears Duration Time", tearsClosingTimeOffset, (float value) => {
				tearsClosingTimeOffset = value;
				settings["General"]["TearsDurationTime"] = value.ToString();
				saveConfig();
			}, 0f, 60f);
			CreateSlider(jfTears, true);
			
			 jfEyecontrol = new JSONStorableFloat("Eyes Control interval Time", eyecontrolClosingTimeInterval, (float value) => {
				eyecontrolClosingTimeInterval = value;
				settings["General"]["EyesControlDurationTime"] = value.ToString();
				saveConfig();
			}, 0f, 60f);
			CreateSlider(jfEyecontrol, true);			
			
			 jfExpressionUp = new JSONStorableFloat("Expression Up Time", expressionClosingTimeOffset1, (float value) => {
				expressionClosingTimeOffset1 = value;
				settings["General"]["ExpressionUpTime"] = value.ToString();
				saveConfig();
			}, 0f, 1f);
			CreateSlider(jfExpressionUp, true);
			
			 jfExpressionDown = new JSONStorableFloat("Expression Down Time", expressionClosingTimeOffset2, (float value) => {
				expressionClosingTimeOffset2 = value;
				settings["General"]["ExpressionDownTime"] = value.ToString();
				saveConfig();
			}, 0f, 60f);
			CreateSlider(jfExpressionDown, true);
			
			blushType = new JSONStorableStringChooser("Blush Type", blushTypes, blushTypeDefault, "Blush Type", (string type) => {
				blushAliasLow = blushMap[type];
				blushTypeDefault = type;
				settings["General"]["DefaultBlushType"] = blushTypeDefault;
				if(blushOnGlobal){
					isBlushingActive = false;
					updateBlush();
				}
				saveConfig();
			});
            RegisterStringChooser(blushType);
            UIDynamicPopup udp = CreateScrollablePopup(blushType, false);
			
			resetButton = CreateButton("Load Default", true);
			resetButton.button.onClick.AddListener(() =>{
				logger("load default settings.");
				loadSettings("default");
				saveConfig();
				removeConfigMenu();
				Start();
			});	

			 toggleDebug = new JSONStorableBool("Debug Message Off", disableLog, (bool toggle) => {
				disableLog = disableLog ? false : true;
				settings["General"]["DebugMessageOff"] = disableLog ? "true" : "false";
				if(disableLog == false){
					logger("disableLog: " + disableLog);
				}
				saveConfig();
			});
			CreateToggle((toggleDebug), false);
			
			 jfThreshold1 = new JSONStorableFloat("Stage1 Threshold", (float)stageThreshold[1], (float value) => {
				stageThreshold[1] = (int)value;
				settings["General"]["StageThreshold1"] = stageThreshold[1].ToString();
				saveConfig();
			}, 0f, 1000f);
			CreateSlider(jfThreshold1, true);			

			 jfThreshold2 = new JSONStorableFloat("Stage2 Threshold", (float)stageThreshold[2], (float value) => {
				stageThreshold[2] = (int)value;
				settings["General"]["StageThreshold2"] = stageThreshold[2].ToString();
				saveConfig();
			}, 0f, 1000f);
			CreateSlider(jfThreshold2, true);			

			 jfThreshold3 = new JSONStorableFloat("Stage3 Threshold", (float)stageThreshold[3], (float value) => {
				stageThreshold[3] = (int)value;
				settings["General"]["StageThreshold3"] = stageThreshold[3].ToString();
				saveConfig();
			}, 0f, 1000f);
			CreateSlider(jfThreshold3, true);			

			 jfThreshold4 = new JSONStorableFloat("Stage4 Threshold", (float)stageThreshold[4], (float value) => {
				stageThreshold[4] = (int)value;
				settings["General"]["StageThreshold4"] = stageThreshold[4].ToString();
				saveConfig();
			}, 0f, 1000f);
			CreateSlider(jfThreshold4, true);
		}
		
		public void removeConfigMenu(){
			RemovePopup(blushType);
			RemoveToggle(toggleAE);
			RemoveToggle(toggleBlush);
			RemoveToggle(toggleTears);
			RemoveToggle(toggleDrool);
			RemoveToggle(toggleTearsGloss);
			RemoveToggle(toggleExpression);
			RemoveToggle(toggleEyecontrol);
			RemoveToggle(toggleEyecontrolRollup);
			RemoveToggle(toggleDebug);
			
			RemoveButton(resetButton);
			
			RemoveSlider(jfBlush);
			RemoveSlider(jfTears);
			RemoveSlider(jfEyecontrol);
			RemoveSlider(jfExpressionUp);
			RemoveSlider(jfExpressionDown);
			RemoveSlider(jfThreshold1);
			RemoveSlider(jfThreshold2);
			RemoveSlider(jfThreshold3);
			RemoveSlider(jfThreshold4);
		}
		
		public void saveConfig(){
			SuperController.singleton.SaveJSON(settings, currentSettingsPath);
		}

		#region eyeControl functions
		public void setInitialTransformOfEyeControl(){
			eyeControl.transform.position = headControl.transform.position;
			eyeControl.transform.rotation = headControl.transform.rotation;
			eyeControl.transform.Translate(0f, 0f, 1f);
		}
		
		public void updateEyecontrol(){
			if(eyecontrolOnGlobal == false){
				return;
			}

			float delta = Time.time - eyecontrolOpeningTime;
			if(delta < eyecontrolClosingTime){
				return;
			}
			float randomOffset = UnityEngine.Random.Range(0.1f, eyecontrolRandomOffset);
			eyecontrolClosingTime = eyecontrolClosingTimeInterval + randomOffset;
			isEyecontrolActive = true;
			
			eyecontrolOpeningTime = Time.time;
			string direction = calcEyeDirection();
			string[] directions = direction.Split(':');
			string upDown = directions[0];
			string rightLeft = directions[1];
			float eyecontrolStrength = calcEyeStrength();
			if(upDown == "neutral" && rightLeft == "neutral"){
				eyeControlBaseSwitch("Player");
			}else{
				changeEyeControl(upDown, rightLeft, eyecontrolStrength);
				eyeControlBaseSwitch("Target");
			}
		}

		public float calcEyeStrength(){
			float strength = 0;
			if(gauge < stageThreshold[1]){
				strength = 0;
			}else if(gauge < stageThreshold[2]){
				strength = UnityEngine.Random.Range(0.1f, 0.2f);
			}else if(gauge < stageThreshold[3]){
				strength = UnityEngine.Random.Range(0.2f, 0.5f);
			}else if(gauge < stageThreshold[4]){
				strength = UnityEngine.Random.Range(0.2f, 0.5f);
			}else{
				if(eyecontrolRollupOnGlobal){
					strength = UnityEngine.Random.Range(0.7f, 1.5f);
				}else{
					strength = UnityEngine.Random.Range(0.2f, 0.5f);
				}
			}
			return strength;
		}
		
		public string calcEyeDirection(){
			string direction = "neutral:neutral";
			
			float rand1 = UnityEngine.Random.Range(0f, 1.0f);
			float rand2 = UnityEngine.Random.Range(0f, 1.0f);
			if(gauge < stageThreshold[1]){
				direction = "neutral:neutral";
			}else if(gauge < stageThreshold[2]){
				if(rand1 < 0.2f){
					direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
				}else{
					direction = "neutral:neutral";
				}
			}else if(gauge < stageThreshold[3]){
				if(rand1 < 0.5f){
					direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
				}else{
					direction = "neutral:neutral";
				}
			}else if(gauge < stageThreshold[4]){
				if(rand1 < 0.8f){
					direction = (rand2 <= 0.5f) ? "down:right" : "down:left";
				}else{
					direction = "neutral:neutral";					
				}
			}else{
				if(rand1 < 0.9f){
					if(eyecontrolRollupOnGlobal){
						direction = "up:neutral";
					}else{
						direction = (rand2 <= 0.5f) ? "down:right" : "down:left";						
					}
				}else{
					direction = "neutral:neutral";					
				}
			}
			
			return direction;
		}
		
		public void eyeControlBaseSwitch(string type){
			atom.GetStorableByID("Eyes").SetStringChooserParamValue("lookMode", type);
			if(type == "Player"){
				setInitialTransformOfEyeControl();
			}
		}
		
		public void manageEyecontrol(float delta){
			if(isEyecontrolActive){
				if(delta > eyecontrolClosingTime){
					eyeControlBaseSwitch("Player");
					isEyecontrolActive = false;
				}
			}
		}

		public void changeEyeControl(string upDown, string rightLeft, float strength){
			setInitialTransformOfEyeControl();
			float degreeUpDown = 0f;
			float degreeRightLeft = 0f;
			if(upDown != "neutral"){
				degreeUpDown = (upDown == "up") ? -1 * baseDegree : baseDegree/2;
				degreeUpDown *= strength;
				eyeControl.transform.RotateAround(
					headControl.transform.position, 
					headControl.transform.right, degreeUpDown); 
			}
			if(rightLeft != "neutral"){
				degreeRightLeft = (rightLeft == "left") ? -1 * baseDegree : baseDegree;	
				degreeRightLeft *= strength;
				eyeControl.transform.RotateAround(
					headControl.transform.position, 
					headControl.transform.up, degreeRightLeft); 
			}
		}
		#endregion

		#region trigger functions
		public void RegisterActions(){
			RegisterActionPair("lNippleTrigger", "OnNippleTouch", "Nipple");
			RegisterActionPair("rNippleTrigger", "OnNippleTouch", "Nipple");
			RegisterActionPair("LabiaTrigger", "OnLabiaTouch", "Labia");
			RegisterActionPair("VaginaTrigger", "OnVaginaTouch", "Vagina");
			RegisterActionPair("DeepVaginaTrigger", "OnVaginaDeepTouch", "VaginaDeep");
			RegisterActionPair("DeeperVaginaTrigger", "OnVaginaDeeperTouch", "VaginaDeeper");
		}
		
		public void RegisterActionPair(string collisionTrigger, string action, string position){
			JSONStorableAction tmpAction = new JSONStorableAction(action, () =>{
				OnErogenousZoneTouch(position);
			});
			RegisterAction(tmpAction);
			createTrigger(collisionTrigger, action);
		}
		
		public void resetTriggers() {
			var table = new Dictionary<string, int>{
				{"OnNippleTouch", 1},
				{"OnLabiaTouch", 1},
				{"OnVaginaTouch", 1},
				{"OnVaginaDeepTouch", 1},
				{"OnVaginaDeeperTouch", 1}
			};
			string[] cTriggerNames = {"rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
			foreach(string cTriggerName in cTriggerNames){
				logger("cTriggerName = " + cTriggerName);
				CollisionTrigger cTrigger = atom.GetStorableByID(cTriggerName) as CollisionTrigger;
				JSONClass triggerClass = cTrigger.trigger.GetJSON();
				JSONArray triggerArray = triggerClass["startActions"].AsArray;
				for (int i=0;i< triggerArray.Count;i++) {
					logger("triggerArray[" + i + "][name]=" + triggerArray[i]["name"].Value);
					if (table.ContainsKey(triggerArray[i]["name"].Value)){
						triggerArray.Remove(i);
					}
				}
				cTrigger.trigger.RestoreFromJSON(triggerClass);
			}
		}
		
		public void createTrigger(string triggerCollisionTriggerName, string scriptTriggerName){
			CollisionTrigger cTrigger = atom.GetStorableByID(triggerCollisionTriggerName) as CollisionTrigger;
			cTrigger.enabled = true;
			TriggerActionDiscrete startTrigger;
			startTrigger=cTrigger.trigger.CreateDiscreteActionStartInternal();
			startTrigger.name = scriptTriggerName;
			startTrigger.receiverAtom = atom;
			startTrigger.receiver = GetPluginStorableById(GetContainingAtom(), myClassName);
			startTrigger.receiverTargetName = scriptTriggerName;
		}
	
		// Function taken from VAMDeluxe's code
		static JSONStorable GetPluginStorableById(Atom atom, string id){
			string storableIdName = atom.GetStorableIDs().FirstOrDefault((string storeId) =>{
				if (string.IsNullOrEmpty(storeId)){
					return false;
				}
				return storeId.Contains(id);
			});
			if (storableIdName == null){
				return null;
			}
			return atom.GetStorableByID(storableIdName);
		}
		#endregion
		
		#region blush functions
		public void switchFaceDecal(string decal){
			logger("new decal:" + decal);
			faceDecalUrl.val = decal;
		}

		public void updateBlush(){
			if(blushOnGlobal == false){
				return;
			}
			blushClosingTime = Time.time + blushClosingTimeOffset;
			if(isBlushingActive == false){
				isBlushingActive = true;
				switchFaceDecal(blushAliasLow);
			}else if(blushLevelUp){
				switchFaceDecal(blushAliasHigh);
				blushLevelUp = false;
			}
		}
		
		public void manageBlush(float currentTime){
			if(isBlushingActive && blushClosingTime < currentTime){
				gauge = 0;
				logger("blush effect stop.");
				switchFaceDecal("");
				isBlushingActive = false;
			}
		}
		
		public void refreshTears(){
			unequipTears();
			tearsString = "";
			droolString = "";
			updateTears();
		}
		#endregion

		#region expression functions
		public void updateExpression(){
			logger("expressionOnGlobal:" + expressionOnGlobal);
			if(expressionOnGlobal == false){
				return;
			}
			float delta = Time.time - expressionOpeningTime;
			float randomOffset = UnityEngine.Random.Range(0.1f,expressionRandomOffset);

			isExpressionActive = true;
			logger("delta:" + delta);
			logger("expressionClosingTimeOffset1:" + expressionClosingTimeOffset1);
			logger("randomOffset:" + randomOffset);
			if(delta < expressionClosingTimeOffset1 + randomOffset){
				return;
			}

			expressionOpeningTime = Time.time;
			targetExpressionStrength = UnityEngine.Random.Range(0.8f,1.0f);
			if(targetExpressionStrength < currentExpressionStrength){
				targetExpressionStrength = currentExpressionStrength + 0.01f;
			}
			expressionPatternOld = expressionPattern;
			expressionPattern = UnityEngine.Random.Range(1, 4);
			initialExpressionStrength = currentExpressionStrength;
			targetEyeClose = eyeCloseMax[expressionPattern] * UnityEngine.Random.Range(0.0f,0.6f);
			initialEyeClose = currentEyeClose;
			logger("currentExpressionStrength:" + currentExpressionStrength);
		}

		public void manageExpression(float delta){
			float blendRatio = calcBlendNewRatio(delta);
			if(isExpressionActive){
				calcExpressionStrength(delta);
				if(currentExpressionStrength == 0){
					isExpressionActive = false;
				}
				if(expressionPatternOld == expressionPattern){
					changeExpressionSingle();
					changeEyeClose(currentEyeClose);
				}else{
					changeExpressionBlend(blendRatio);
					changeEyeClose(currentEyeClose);						
				}
			}
		}
		
		public void calcExpressionStrength(float delta){
			if(delta < expressionClosingTimeOffset1){
				currentExpressionStrength = (targetExpressionStrength - initialExpressionStrength) / expressionClosingTimeOffset1 * delta + initialExpressionStrength;
				currentEyeClose = (targetEyeClose - initialEyeClose) / expressionClosingTimeOffset1 * delta + initialEyeClose;
			}else if(delta < expressionClosingTimeOffset2){
				currentExpressionStrength = targetExpressionStrength - (targetExpressionStrength/expressionClosingTimeOffset2) * (delta - expressionClosingTimeOffset1);
				currentEyeClose = targetEyeClose - (targetEyeClose/expressionClosingTimeOffset2) * (delta - expressionClosingTimeOffset1);						
			}else{
				currentExpressionStrength = 0f;
				currentEyeClose = 0f;
			}
		}
		
		public float calcBlendNewRatio(float delta){
			float ratio = delta / expressionClosingTimeOffset1;
			if(ratio > 1){
				ratio = 1;
			}
			return ratio;
		}
		
		public void changeExpressionSingle(){
			for(int i = 0; i < 4; i++){
				if(i == expressionPattern){
					Emorph[i].SetValue(currentExpressionStrength);
				}else{
					Emorph[i].SetValue(0f);					
				}
			}
		}
		
		public void changeExpressionBlend(float ratio){
			float blendNew = ratio;
			float blendOld = 1 - ratio;
			for(int i = 0; i < 4; i++){
				if(i == expressionPattern){
					Emorph[i].SetValue(currentExpressionStrength * blendNew);
				}else if(i == expressionPatternOld){
					Emorph[i].SetValue(currentExpressionStrength * blendOld);					
				}else{
					Emorph[i].SetValue(0f);					
				}
			}
		}	
		
		public void stopExpression(){
			for(int i = 0; i < 4; i++){
				Emorph[i].SetValue(0f);					
			}
			EmorphEyesClosed.SetValue(0f);
		}
		
		public void changeEyeClose(float strength){
			EmorphEyesClosed.SetValue(strength);
		}
		#endregion
		
		#region tears functions
		public void updateTears(){
			if(tearsOnGlobal == false){
				return;
			}
			tearsClosingTime = Time.time + tearsClosingTimeOffset;

			string lastTearsString = tearsString;
			string lastDroolString = droolString;
			isTearsUpdated = false;
			tearsString = "Tears ";
			droolString = "Drool";
			droolOn = false;
			tearsOn = false;
			
			if(gauge < stageThreshold[1]){
				tearsOn = false;
			}else if(gauge < stageThreshold[2]){
				tearsOn = true;
				tearsString += "Small";
			}else if(gauge < stageThreshold[3]){
				tearsOn = true;
				tearsString += "Middle";				
			}else if(gauge < stageThreshold[4]){
				tearsOn = true;
				tearsString += "Large";
			}else if(gauge >= stageThreshold[4]){
				tearsString += "Large";
				tearsOn = true;
				droolOn = droolOnGlobal ? true : false;
//				blushLevelUp = true;
			}
			
			if(isTearsGloss){
				tearsString += "(Gloss)";
				droolString += "(Gloss)";
			}

			if(tearsOn){
				geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], true);
				isTearsActive = true;
			}
			if(droolOn){
				geometryForClothes.SetActiveClothingItem(tearsMap[droolString], true);								
			}
			if(tearsString != lastTearsString){
				isTearsUpdated = true;
				geometryForClothes.SetActiveClothingItem(tearsMap[lastTearsString], active: false);
			}else if(droolString != lastDroolString){
				isTearsUpdated = true;
				geometryForClothes.SetActiveClothingItem(tearsMap[lastDroolString], active: false);
			}	
		}
		
		public void unequipTears(){
			if(tearsOn){
				geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], active: false);
			}
			if(droolOn){
				geometryForClothes.SetActiveClothingItem(tearsMap[droolString], active: false);				
			}
		}
		
		public void manageTears(float currentTime){
			if(isTearsActive && tearsClosingTime < currentTime){
				gauge = 0;
				logger("tears effect stop.");
				isTearsActive = false;
				geometryForClothes.SetActiveClothingItem(tearsMap[tearsString], active: false);
				geometryForClothes.SetActiveClothingItem(tearsMap[droolString], active: false);
			}
		}
		#endregion
		
		#region engine
		public void OnErogenousZoneTouch(string zone){ 
			logger("Touch: " + zone);
			updateExpression();
			updateBlush();
			updateTears();
			updateEyecontrol();
			
			gauge++;
		}
		
        public void Update(){
            try{
				if(Time.time > currentTime + 10){
					logger("start plug-in or load scene.");
					maintenanceSequence();
				}

				currentTime = Time.time;
				float delta = currentTime - expressionOpeningTime;

				manageBlush(currentTime);
				manageTears(currentTime);
				manageExpression(delta);
				manageEyecontrol(delta);
			}catch(Exception e){
//				SuperController.LogError("Failed in Update: " + e);
			}
		}
		#endregion
		
		#region etc
		public void logger(string str){
			if(disableLog == false){
				SuperController.LogMessage(str);
			}
		}
		
		// Function taken from VAMDeluxe's code
		public string GetPluginPath(){
			string id = name.Substring(0, name.IndexOf('_'));
            string filename = manager.GetJSON()["plugins"][id].Value;
            return filename.Substring(0, filename.LastIndexOfAny(new char[] { '/', '\\' }));
		}
		#endregion
	}
}
